# 从.warc或.warc.gz中提取指定mime类型的文件
# 逐一读取warc_path中的所有warc/warc.gz文件，提取指定文件放到output_dir指定的文件夹中，
# 当前指定的mime类型是：'application/pdf', 'text/html', 'text/plain', 'text/markdown'
# 错误日志位于output_dir中的_errors.log
# 提取的文件会在文件名前添加域名，例如www.aaa.com_index.html
# 如果有重名，则自动添加序号，例如例如www.aaa.com_index_99.html，这意味着如果多次运行脚本，会在output_dir中生成多个内容相同但名字不同的文件。
# 
# pip install warcio
# 已知问题：
# warc包中的某些文件可能不正常，例如一个网页中连接的图片1.gif在源服务器上已不存在，
# 服务器返回了一个自定义错误页面而不是http标准错误，那么爬虫可能在磁盘上保存一个名为1.gif内容是http err文本的文件，
# 这个文件在warc元数据中的文件类型是html/text类型


import os  
import warcio
import logging  
from warcio.archiveiterator import ArchiveIterator
from datetime import datetime  
from urllib.parse import urlparse

# 判断是否有重名文件，如有则在文件名后加_数字序号
def generate_unique_filename(filename, output_dir):
    base, ext = os.path.splitext(filename)  
    counter = 1
    unique_filename = filename
    while os.path.exists(os.path.join(output_dir, unique_filename)):  
        base_with_counter = f"{base}_{counter}"
        unique_filename = f"{base_with_counter}{ext}"
        counter += 1
    return os.path.join(output_dir, unique_filename) 

def extract_files_from_warc(warc_path, output_dir): 
    # 确保输出目录存在  
    os.makedirs(output_dir, exist_ok=True)  
    # 配置日志记录 
    log_file_path = os.path.join(output_dir, '_errors.log')
    logging.basicConfig(filename=log_file_path, level=logging.ERROR)  

    error_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    print(error_time , "=====开始新一轮warc处理=====")
    logging.error(error_time + "=====开始新一轮warc处理=====")
    
    # 常见的文本文件类型，已废弃，改为使用mime类型  
    # text_file_types = ('.html', '.txt', '.pdf')  
    
    # 遍历warc_path下的每一个子文件夹  
    for subdir, dirs, files in os.walk(warc_path):  
        for file in files:  
            if file.endswith('.warc') or file.endswith('.warc.gz'):
                warc_path = os.path.join(subdir, file)  
                try:  
                    # 读取warc文件
                    with open(warc_path, 'rb') as stream:
                        reader = warcio.ArchiveIterator(stream)
                        for record in reader:  
                            if record.rec_type == 'response': 
                                # http response不是200的直接跳过
                                if not record.http_headers.statusline.startswith("200"):
                                    continue
                                content_type = record.http_headers.get_header('Content-Type')
                                if content_type.split(';')[0].strip() in ['application/pdf', 'text/html', 'text/plain', 'text/markdown']:  
                                    # WARC-Target-URI是类似于https://www.aaa.com/bbb/ccc.pdf的字符串
                                    original_file_URI = record.rec_headers.get_header('WARC-Target-URI')
                                    # 跳过robots.txt
                                    if 'robots.txt' in original_file_URI:
                                        continue
                                    # 从URI提取路径，/bbb/ccc.pdf
                                    parsed_uri = urlparse(original_file_URI)
                                    filename = parsed_uri.path
                                    # 从路径中提取文件名，ccc.pdf，如果是根路径，则提取空字符串
                                    filename = os.path.basename(filename.rstrip('/'))
                                    # 如果是根目录https://www.aaa.com/，则将文件名定为index.html
                                    if filename == "":
                                        filename = "index.html"
                                    # 在文件名前面附加网站域名www.aaa.com_
                                    domain = parsed_uri.netloc
                                    if ':' in domain:  
                                        domain = domain.split(':')[0]
                                    filename = domain + "_" + filename
                                    # 检测output_dir中是否存在同名文件，如果有，则在末尾添加_数字序号
                                    filename = generate_unique_filename(filename, output_dir)
                                    # 写入文件内容  
                                    with open(filename, 'wb') as f_out:  
                                        f_out.write(record.content_stream().read())  
    
                except Exception as e:  
                    # 错误处理  
                    error_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")  
                    error_message = f" Error processing {warc_path} : {e}"  
                    print(error_time, error_message)  
                    logging.error(error_time + error_message)

    error_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    print(error_time , "=====结束新一轮warc处理=====")
    logging.error(error_time + "=====结束新一轮warc处理=====")

# 设置输入和输出目录  
warc_path = r'E:\mnbvc\warc\test'  # WARC和CDX文件所在的根目录  
output_dir = r'E:\mnbvc\warc\output_warcExtract'  # 输出目录  
extract_files_from_warc(warc_path, output_dir)